/*
  ggHiggs code based on

    R. D. Ball, M. Bonvini, S. Forte, S. Marzani, G. Ridolfi
    arXiv:1303.3590

  please cite it if you use it
*/

#include <iomanip>
#include <vector>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <stdio.h>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/stat.h>
#include <sys/time.h>
#include <getopt.h>


#include "./src/ggHiggs.hh"
#include "LHAPDF/LHAPDF.h"  // interface with LHAPDF
#include "LHAPDF/PDFSet.h"  // interface with LHAPDF



using namespace std;
using namespace ggHiggs;


/*
  If you like, you can customize your alpha_s routine.
  However, consistency with PDF sets suggests the use of the LHAPDF alphas routine.
*/
double ggHiggs::alphas(double mu) {
  return alphasPDF(mu);
  //return alphas_ihixs(mu, 3, alphasPDF(GetZmass()), GetZmass()); // taken from ihixs code
  //return alphas_ihixs(mu, order, alphasPDF(GetZmass()), GetZmass()); // taken from ihixs code
}


void print_usage() {
  cout << endl
       << "Usage:" << endl
       << " -p  --PDF-set <set>    specify PDFs being used (default \"PDF4LHC15_nnlo_100\")" << endl
       << " -m  --mass <mass>      set Higgs mass [GeV]" << endl
       << " -E  --Ecm <energy>     set collider energy [GeV]" << endl
       << " -f  --fact <murat>     set muF/mH ratio" << endl
       << " -r  --ren <murat>      set muR/mH ratio" << endl
       << " -t  --topmass <mass>   set top quark mass [GeV] (either mt_pole or mt(mt), depending on the -S flag)" << endl
       << " -b  --include-bottom   include the bottom Yukawa coupling" << endl
       << " -c  --include-charm    include the charm  Yukawa coupling" << endl
       << " -e  --noEFT            use finite quark masses (default: infinite-mtop EFT approximation)" << endl
#ifdef withHELL
       << " -x  --small-x          include small-x resummation (from HELL-x) in the ggH partonic cross sections" << endl
#endif
       << " -S  --MSbarMass        use MSbar running top mass (with mt(mt) = 162.7 GeV)" << endl
       << " -o  --CPodd            pseudo-scalar (CP odd) Higgs production" << endl
       << " -a  --accuracy         fixed order accuracy" << endl
       << " -v  --verbose          show verbose output" << endl
       << endl;
  cout << "Most realistic setup:" << endl
       << "  ggHiggs -e -b -c" << endl
       << endl;
  exit(0);
  return;
}




double muFrat = 1;
double muRrat = 1;
void read_arguments(int argc, char* argv[], double &mH, double &sqrts, string &set, bool &nob, bool &noc, bool &smallx, int &accuracy) {
  const char* const short_options = "ht:ep:f:r:m:E:bcSvoxa:";
  const struct option long_options[] = { { "help", 0, NULL, 'h' },
					 { "topmass", 1, NULL, 't' },
					 { "noEFT", 0, NULL, 'e' },
					 { "PDF-set", 1, NULL, 'p' },
					 { "fact", 1, NULL, 'f' },
					 { "ren", 1, NULL, 'r' },
					 { "mass", 1, NULL, 'm' },
					 { "Ecm", 1, NULL, 'E' },
					 { "include-bottom", 0, NULL, 'b' },
					 { "include-charm", 0, NULL, 'c' },
					 { "MSbarMass", 0, NULL, 'S' },
					 { "CPodd", 0, NULL, 'o' },
					 { "small-x", 0, NULL, 'x' },
					 { "accuracy", 1, NULL, 'a' },
					 { "verbose", 0, NULL, 'v' },
					 { NULL, 0, NULL, 0 } };
  int next_option;
  ostringstream sset, sa;
  do {
    next_option = getopt_long (argc, argv, short_options, long_options, NULL);
    switch (next_option) {
    case 'h':
      print_usage();
    case 'e':
      finite_mt = true;
      break;
    case 't':
      mtop = strtod(optarg, NULL);
      mtmt = mtop;
      break;
    case 'p':
      sset << optarg;
      set = sset.str();
      break;
    case 'f':
      muFrat = strtod(optarg, NULL);
      break;
    case 'r':
      muRrat = strtod(optarg, NULL);
      break;
    case 'm':
      mH = strtod(optarg, NULL);
      break;
    case 'E':
      sqrts = strtod(optarg, NULL);
      break;
    case 'b':
      nob = false;
      break;
    case 'c':
      noc = false;
      break;
    case 'S':
      MSbarMass = true;
      cout << "\033[0;31m WARNING: MSbar scheme for the top mass is a new feature, implemented only for the top in the large-mt EFT at the moment. Use with care.\033[0m" << endl;
      break;
    case 'o':
      SetCPcharge(CPodd);
      break;
    case 'x':
      smallx = true;
      break;
    case 'a':
      accuracy = strtoul(optarg, NULL, 10);
      break;
    case 'v':
      quiet = false;
      break;
    case '?':
      print_usage();
    case -1: break;
    default: abort();
    }
  }
  while (next_option != -1);
  //
  return;
}

int main (int argc, char* argv[]) {

  // relative precision for integrals
  INTEGRAL_PRECISION = 5.e-3;

  // default values (modified by runtime arguments)
  double mH = 125;
  double sqrts = 13000;
  string setname = "PDF4LHC15_nnlo_mc";
  int accuracy = 3;

  // read runtime arguments
  bool nob=true, noc=true; // when true, bottom and charm do NOT run into the loop coupling to the Higgs
  bool smallxRes = false;
  read_arguments(argc, argv, mH, sqrts, setname, nob, noc, smallxRes, accuracy);
  _muF_mu_ratio  = muFrat;
  _muR_muF_ratio = muRrat/muFrat;
#ifdef withHELL
  if(smallxRes && !finite_mt) {
    cout << "\033[0;31m  Warning: Small-x resummation assumes exact theory, so it's not available within the EFT. Try running with -e -x\033[0m  " << endl;
    smallxRes = false;
  }
#endif

  // Start time counter
  struct timeval t0, t1;
  gettimeofday(&t0,NULL);

  // init PDFs
  cout << "Loading PDF set " << setname << endl;
  int subset = 0;
  pdf_init(setname, subset);
  // init LHAPDF6 set (necessary to compute the error)
  const LHAPDF::PDFSet set(setname);
  int nmembers = set.size();

  // masses (negative mass means the quark will not contribute to the loop)
  mbot = (nob ? -1 : 1) * getQuarkMass(5);
  mcha = (noc ? -1 : 1) * getQuarkMass(4);
  // init
  double tau = mH*mH/sqrts/sqrts;
  sxc = new SmallxCoeffs(mH,mtop,mbot,mcha);
  if(smallxRes) SetSmallxResummation(true);
  init_consts(mH);
  SetSmallxMode(0,0);

 /******* RUN EXAMPLE *******/

  // Note also that the same PDF set has to be used at every order.
  double LO, NLO, NNLO, NNNLO;
  double LOpref = xs_LO_prefactor(mH);
  double muF = mH *_muF_mu_ratio;
  double muR = muF*_muR_muF_ratio;
  double alpha_s = alphas(muR);

  map<int,string> forder;
  forder.insert(pair<int,string>(0,"LO"));
  forder.insert(pair<int,string>(1,"NLO"));
  forder.insert(pair<int,string>(2,"NNLO"));
  forder.insert(pair<int,string>(3,"NNNLO"));

  // results
  double rexact[3];

  bool ggonly = false;

  vector<double> sigma;

  if(accuracy>=4)
  {
    cout << "Please choose a value between 0 (LO) and 3 (NNNLO)" << endl;
    return 1;
  }

  for (int j = 0; j < nmembers ; ++j)
  {
    LO = 0;
    NLO = 0;
    NNLO = 0;
    NNNLO = 0;

    // init PDF member
    pdf_init(setname, j);

    // LO
    LO = tau*Lum(tau,muF)*LOpref*pow(alpha_s,2);

    // NLO
    if(accuracy >= 1)
    {
      order = 1;
      __all_channels__ = true & !ggonly;
      compute(mH, sqrts, LO, 0, 0, &rexact[0]);
      NLO = LO + rexact[0];
      if(smallxRes)
        NLO   += computeSxRes(mH, sqrts, LO, 1);
    }

    // NNLO
    if(accuracy >=2)
    {
      order = 2;
      __all_channels__ = true & !ggonly;
      compute(mH, sqrts, LO, NLO, 0, &rexact[1]);
      NNLO = NLO + rexact[1];
      if(smallxRes)
          NNLO  += computeSxRes(mH, sqrts, LO, 2);
    }

    // NNNLO
    if(accuracy >=3)
    {
      order = 3;
      //__all_channels__ = false;
      //compute(mH, sqrts, LO, NLO, NNLO, NULL, &rsoft2[2], &rsoft1[2], &rNsoft[2], &rhighenergy[2], &momconsunc[2], &rhighenergyLL[2]);
      __all_channels__ = true & !ggonly;
      compute(mH, sqrts, LO, NLO, NNLO, &rexact[2]);
      NNNLO = NNLO + rexact[2];
      if(smallxRes)
        NNNLO += computeSxRes(mH, sqrts, LO, 3);
    }

    if( accuracy==0 )
      sigma.push_back(LO);
    else if( accuracy==1  )
      sigma.push_back(NLO);
    else if( accuracy==2  )
      sigma.push_back(NNLO);
    else if( accuracy==3  )
      sigma.push_back(NNNLO);
    else
    {
      cout << "Unknown accuracy; please choose a value between 0 (LO) and 3 (NNNLO)" << endl;
      return 1;
    }


  }

  //
  cout << "Higgs mass:       " << mH << " GeV" << endl
       << "Collider energy:  " << sqrts << " GeV" << endl
       << "tau = mH^2/s =    " << tau << endl
       << "top    mass =     " << mtop << "  running in the loop" << endl
       << "bottom mass =     " << fabs(mbot) << "  " << (nob ? "NOT " : "") << "running in the loop" << endl
       << "charm  mass =     " << fabs(mcha) << "  " << (noc ? "NOT " : "") << "running in the loop" << endl
       << "quark mass effects beyond LO: " << (finite_mt ? "YES (exact theory)" : "NO (EFT)") << endl
       << "Renormalization scale muR = " << _muR_muF_ratio*_muF_mu_ratio << " * mH" << endl
       << "Factorization   scale muF = " << _muF_mu_ratio << " * mH" << endl
       << "alphas(muR) = " << alphas(_muR_muF_ratio*_muF_mu_ratio*mH) << endl;
  //


  const LHAPDF::PDFUncertainty unc = set.uncertainty(sigma,-1);
  double central  = unc.central;
  double errplus  = unc.errplus;
  double errminus = unc.errminus;

  cout  << endl << endl << endl;

  cout  << "Total cross-section at " << forder.at(accuracy) << ":" << endl
        << central << " + " << errplus << " - " << errminus
        << endl << endl;

  // Stop time counter
  gettimeofday(&t1,NULL);
  cout << "Total time: " << t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001 << "s" << endl;


  return 0;

}
